using System;
using System.Collections.Generic;
using UnityEngine;
using System.Linq;
using System.Collections;
using MVR.FileManagementSecure;

namespace ToumeiHitsuji
{

    public partial class SlapStuff
    {
        public partial class SlapCollisionController : JSONStorableParamGroup
        {
            protected SlapStuff _parentScript = null;

            private JSONStorableFloat _velocityStartJSON;
            private JSONStorableFloat _velocityEndJSON;
            private JSONStorableBool _cutoffAtEndVelocityJSON;
            private JSONStorableBool _useFrontalVelocityJSON;
            private JSONStorableFloat _volumeStartJSON;
            private JSONStorableFloat _volumeEndJSON;
            private JSONStorableFloat _pitchStartJSON;
            private JSONStorableFloat _pitchEndJSON;
            private JSONStorableFloat _interpolationExponentJSON;

            private JSONStorableStringList _colliderListJSON;
            private JSONStorableString _colliderListTextJSON;
            private JSONStorableStringChooser _colliderChooser;
            private JSONStorableBool _showAllCollidersJSON;
            private JSONStorableStringList _targetColliderFilterJSON;
            private JSONStorableStringChooser _targetColliderFilterChooser;
            private JSONStorableFloat _timeOutJSON;
            private JSONStorableStringChooser _atomTypeChooser;
            private JSONStorableBool _preventSelfCollisionJSON;
            private List<SlapCollisionHandler> _collisionHandlers = new List<SlapCollisionHandler>();

            private JSONStorableStringList _audioUrlListJSON;
            private List<NamedAudioClip> _audioClips = new List<NamedAudioClip>();
            private JSONStorableString _audioUrlListTextJSON;
            private JSONStorableStringChooser _audioUrlsChooser;
            private JSONStorableUrl _addUrl;
            private JSONStorableString _audioSourceUidJSON;
            private Atom _audioSource = null;

            protected UIDynamicContainer _uiContainerSimple;
            protected UIDynamicContainer _uiContainerAdvanced;
            private bool _creatingAudioSource = false;

            public Atom SlapAudioSource
            {
                get
                {
                    ensureAudioSource();
                    return _audioSource;
                }
            }

            public string RandomId
            {
                get { return UnityEngine.Random.Range(0,0xFFFFFF).ToString("X6"); }
            }

            public string RandomAudioSourceName
            {
                get { return $"SlapAudio#{RandomId}#{_parentScript.containingAtom.name}"; }
            }

            public SlapCollisionController(SlapStuff parentScript)
            {
                _parentScript = parentScript;
                _uiContainerSimple = new UIDynamicContainer(parentScript);
                _uiContainerAdvanced = new UIDynamicContainer(parentScript);

                initBasicSettings();
                initBasicSettingsAdvanced();

                _uiContainerSimple.Add(_parentScript.CreateSpacer(false)).height = 80f;
                _uiContainerSimple.Add(_parentScript.CreateSpacer(true)).height = 80f;

                initColliders();
                initCollidersAdvanced();

                _uiContainerSimple.Add(_parentScript.CreateSpacer(false)).height = 80f;
                _uiContainerSimple.Add(_parentScript.CreateSpacer(true)).height = 80f;

                initAudio();
                initAudioAdvanced();
            }

            private void initBasicSettings()
            {
                initBasicSettingsHelp();

                _velocityStartJSON = new JSONStorableFloat($"(Start) Velocity", 0.5f, 0f, 10f, false);
                _velocityStartJSON.storeType = JSONStorableParam.StoreType.Full;
                RegisterFloat(_velocityStartJSON);
                _uiContainerSimple.Add(_parentScript.CreateSlider(_velocityStartJSON, false));

                _velocityEndJSON = new JSONStorableFloat($"(End) Velocity", 5f, 0f, 10f, false);
                _velocityEndJSON.storeType = JSONStorableParam.StoreType.Full;
                RegisterFloat(_velocityEndJSON);
                _uiContainerSimple.Add(_parentScript.CreateSlider(_velocityEndJSON, true));

                _volumeStartJSON = new JSONStorableFloat($"(Start) Volume", 0f, 0f, 1f, true);
                _volumeStartJSON.storeType = JSONStorableParam.StoreType.Full;
                RegisterFloat(_volumeStartJSON);
                _uiContainerSimple.Add(_parentScript.CreateSlider(_volumeStartJSON, false));

                _volumeEndJSON = new JSONStorableFloat($"(End) Volume", 1f, 0f, 1f, true);
                _volumeEndJSON.storeType = JSONStorableParam.StoreType.Full;
                RegisterFloat(_volumeEndJSON);
                _uiContainerSimple.Add(_parentScript.CreateSlider(_volumeEndJSON, true));

                _pitchStartJSON = new JSONStorableFloat($"(Start) Pitch", 0.75f, -3f, 3f, true);
                _pitchStartJSON.storeType = JSONStorableParam.StoreType.Full;
                RegisterFloat(_pitchStartJSON);
                _uiContainerSimple.Add(_parentScript.CreateSlider(_pitchStartJSON, false));

                _pitchEndJSON = new JSONStorableFloat($"(End) Pitch", 1.25f, -3f, 3f, true);
                _pitchEndJSON.storeType = JSONStorableParam.StoreType.Full;
                RegisterFloat(_pitchEndJSON);
                _uiContainerSimple.Add(_parentScript.CreateSlider(_pitchEndJSON, true));
            }

            private void initBasicSettingsAdvanced()
            {
                _uiContainerAdvanced.Add(_parentScript.CreateSpacer(false)).height = 20f;
                _uiContainerAdvanced.Add(_parentScript.CreateSpacer(true)).height = 25f;

                _useFrontalVelocityJSON = new JSONStorableBool("Use Frontal Velocity Fraction", true);
                _useFrontalVelocityJSON.storeType = JSONStorableParam.StoreType.Full;
                RegisterBool(_useFrontalVelocityJSON);
                _uiContainerAdvanced.Add(_parentScript.CreateToggle(_useFrontalVelocityJSON, true));

                _cutoffAtEndVelocityJSON = new JSONStorableBool("Cutoff At End Velocity", false);
                _cutoffAtEndVelocityJSON.storeType = JSONStorableParam.StoreType.Full;
                RegisterBool(_cutoffAtEndVelocityJSON);
                _uiContainerAdvanced.Add(_parentScript.CreateToggle(_cutoffAtEndVelocityJSON, true));

                _interpolationExponentJSON  = new JSONStorableFloat($"Interpolation Exponent", 2f, 0f, 4f, true);
                _interpolationExponentJSON.storeType = JSONStorableParam.StoreType.Full;
                RegisterFloat(_interpolationExponentJSON);
                _uiContainerAdvanced.Add(_parentScript.CreateSlider(_interpolationExponentJSON, false));
            }

            private void initColliders()
            {
                initCollidersHelp();

                _colliderListJSON = new JSONStorableStringList("Colliders") {
                    setCallbackFunction = (List<string> value) => parseColliders(),
                    addCallbackFunction = (string value) => addCollider(value),
                    removeCallbackFunction = (string value, int index) => removeCollider(index),
                    clearCallbackFunction = clearColliders,
                    storeType = JSONStorableParam.StoreType.Full,
                    restoreTime = JSONStorableParam.RestoreTime.Late
                };
                RegisterParam(_colliderListJSON);

                _colliderListTextJSON = new JSONStorableString($"CollidersText", "<b>Added Colliders (This Atom):</b>");
                _uiContainerSimple.Add(_parentScript.CreateTextField(_colliderListTextJSON, true)).height = 315f;

                _colliderChooser = new JSONStorableStringChooser($"Collider",
                    new List<string>{"None"}, "None", "Collider");

                UIDynamicButton addColliderButton = _uiContainerSimple.Add(_parentScript.CreateButton("Add Collider", false));
                addColliderButton.buttonColor = _parentScript.Green;
                addColliderButton.button.onClick.AddListener(() =>
                {
                    if (string.IsNullOrEmpty(_colliderChooser.val) || _colliderChooser.val == "None" || _colliderListJSON.valNoCallback.Contains(_colliderChooser.val))
                    {
                        Logger.LogDebugMessage($"Can't add collider: {_colliderChooser.val}.");
                    }
                    else
                    {
                        _colliderListJSON.Add(_colliderChooser.val);
                    }
                });

                UIDynamicButton removeColliderButton = _uiContainerSimple.Add(_parentScript.CreateButton("Remove Collider", false));
                removeColliderButton.buttonColor = _parentScript.Red;
                removeColliderButton.button.onClick.AddListener(() =>
                {
                    int indexToRemove = _colliderListJSON.valNoCallback.IndexOf(_colliderChooser.val);
                    if (_colliderListJSON.RemoveAt(indexToRemove))
                    {
                        if (indexToRemove >= _colliderListJSON.valNoCallback.Count)
                        {
                            indexToRemove -= 1;
                        }
                        indexToRemove = indexToRemove < 0 ? 0 : indexToRemove;
                        _colliderChooser.valNoCallback = _colliderChooser.choices[indexToRemove];
                    }
                });

                _uiContainerSimple.Add(_parentScript.CreateFilterablePopup(_colliderChooser, false));
                _showAllCollidersJSON = new JSONStorableBool("Show All Colliders", false, (bool isChecked) => syncColliderChoices());
                _uiContainerSimple.Add(_parentScript.CreateToggle(_showAllCollidersJSON));
                syncColliderChoices();
            }

            private void initCollidersAdvanced()
            {
                _uiContainerAdvanced.Add(_parentScript.CreateSpacer(true)).height = 80f;
                _uiContainerAdvanced.Add(_parentScript.CreateSpacer(false)).height = 80f;

                initConstraintsHelp();

                _targetColliderFilterJSON = new JSONStorableStringList("Target Collider Filter") {
                    setCallbackFunction = (List<string> value) => syncTargetColliderFilterChoices(),
                    addCallbackFunction = (string value) => syncTargetColliderFilterChoices(),
                    removeCallbackFunction = (string value, int index) => syncTargetColliderFilterChoices(),
                    clearCallbackFunction = syncTargetColliderFilterChoices,
                    storeType = JSONStorableParam.StoreType.Full,
                    restoreTime = JSONStorableParam.RestoreTime.Late
                };
                RegisterParam(_targetColliderFilterJSON);

                UIDynamicButton addColliderFilterButton = _uiContainerAdvanced.Add(_parentScript.CreateButton("Add Filter", false));
                addColliderFilterButton.buttonColor = _parentScript.Green;
                UIDynamicButton removeColliderFilterButton = _uiContainerAdvanced.Add(_parentScript.CreateButton("Remove Filter", false));
                removeColliderFilterButton.buttonColor = _parentScript.Red;
                _targetColliderFilterChooser = new JSONStorableStringChooser("TargetColliderFilter", new List<string>(), "", "Target Collider Filter");
                UIDynamicPopup colliderFilterPopup = _uiContainerAdvanced.Add(_parentScript.CreateFilterablePopup(_targetColliderFilterChooser, true));

                addColliderFilterButton.button.onClick.AddListener(() =>
                {
                    if (string.IsNullOrEmpty(colliderFilterPopup.popup.filter))
                    {
                        Logger.LogMessage("Use the \"Filter...\" TextField to type in what you want to add.");
                    }
                    else if (!_targetColliderFilterJSON.valNoCallback.Contains(colliderFilterPopup.popup.filter))
                    {
                        _targetColliderFilterJSON.Add(colliderFilterPopup.popup.filter);
                        _targetColliderFilterChooser.val = colliderFilterPopup.popup.filter;
                    }
                });

                removeColliderFilterButton.button.onClick.AddListener(() =>
                {
                    _targetColliderFilterJSON.Remove(_targetColliderFilterChooser.val);
                    _targetColliderFilterChooser.val = "";
                });

                _uiContainerAdvanced.Add(_parentScript.CreateSpacer(true)).height = 20f;
                _uiContainerAdvanced.Add(_parentScript.CreateSpacer(false)).height = 20f;

                _atomTypeChooser = new JSONStorableStringChooser("Atom Type",
                    new List<string>{"All"}, "All", "Target Atom Type");
                _atomTypeChooser.storeType = JSONStorableParam.StoreType.Full;
                RegisterStringChooser(_atomTypeChooser);
                UIDynamicPopup atomTypePopup = _uiContainerAdvanced.Add(_parentScript.CreateFilterablePopup(_atomTypeChooser, true));
                atomTypePopup.popup.onOpenPopupHandlers += () =>
                {
                    HashSet<string> choicesSet = new HashSet<string>(from atom in SuperController.singleton.GetAtoms() select atom.type);
                    List<string> choicesList = choicesSet.ToList();
                    choicesList.Sort();
                    choicesList.Insert(0, "All");
                    _atomTypeChooser.choices = choicesList;
                };

                _preventSelfCollisionJSON = new JSONStorableBool("Prevent Self Collision", false);
                _preventSelfCollisionJSON.storeType = JSONStorableParam.StoreType.Full;
                RegisterBool(_preventSelfCollisionJSON);
                _uiContainerAdvanced.Add(_parentScript.CreateToggle(_preventSelfCollisionJSON, true));

                _timeOutJSON = new JSONStorableFloat($"Timeout", 0.25f, 0f, 1f, false);
                _timeOutJSON.storeType = JSONStorableParam.StoreType.Full;
                RegisterFloat(_timeOutJSON);
                _uiContainerAdvanced.Add(_parentScript.CreateSlider(_timeOutJSON, false));

                _uiContainerAdvanced.Add(_parentScript.CreateSpacer(false)).height = 55f;
            }

            private void initAudio()
            {
                initAudioHelp();

                _audioSourceUidJSON = new JSONStorableString($"Collision AudioSource", RandomAudioSourceName);
                _audioSourceUidJSON.storeType = JSONStorableParam.StoreType.Full;
                RegisterString(_audioSourceUidJSON);

                _audioUrlListJSON = new JSONStorableStringList("Audio Urls")
                {
                    setCallbackFunction = (List<string> value) => parseAudioUrls(),
                    addCallbackFunction = (string value) => addAudioFile(value),
                    removeCallbackFunction = (string value, int index) => removeAudioFile(index),
                    clearCallbackFunction = clearAudioClips,
                    storeType = JSONStorableParam.StoreType.Full,
                    restoreTime = JSONStorableParam.RestoreTime.Late
                };
                RegisterParam(_audioUrlListJSON);

                _audioUrlsChooser = new JSONStorableStringChooser($"file", new List<string>(), "", "File");
                _audioUrlListTextJSON = new JSONStorableString($"Audio Urls Text", "<b>Added Audio Files:</b>");
                _uiContainerSimple.Add(_parentScript.CreateTextField(_audioUrlListTextJSON, true)).height = 800f;

                UIDynamicButton addAudioFileButton = _uiContainerSimple.Add(_parentScript.CreateButton("Add Audio File", false));
                addAudioFileButton.buttonColor = _parentScript.Green;
                _addUrl = new JSONStorableUrl($"url", "", "wav|mp3|ogg|aif|xm|mod|it|s3m");
                _addUrl.RegisterFileBrowseButton(addAudioFileButton.button);
                _addUrl.setCallbackFunction = (string url) => {
                    if (string.IsNullOrEmpty(url) || _audioUrlListJSON.valNoCallback.Contains(url) || !FileManagerSecure.FileExists(url))
                    {
                        Logger.LogDebugMessage($"Can't add audio url: {url}.");
                    }
                    else
                    {
                        _audioUrlListJSON.Add(url);
                        _audioUrlsChooser.SetVal(_audioUrlsChooser.choices.Last());
                        _addUrl.valNoCallback = "";
                    }
                };

                UIDynamicButton removeAudioFileButton = _uiContainerSimple.Add(_parentScript.CreateButton("Remove Audio File", false));
                removeAudioFileButton.buttonColor = _parentScript.Red;
                removeAudioFileButton.button.onClick.AddListener(() =>
                {
                    if (!string.IsNullOrEmpty(_audioUrlsChooser.val))
                    {
                        //int audioIndex = int.Parse(_audioUrlsChooser.val.Substring(0, _audioUrlsChooser.val.IndexOf(':')));
                        int audioIndex = _audioUrlsChooser.choices.IndexOf(_audioUrlsChooser.val);
                        _audioUrlListJSON.RemoveAt(audioIndex);
                        if (audioIndex >= _audioUrlsChooser.choices.Count)
                        {
                            audioIndex -= 1;
                        }
                        if (audioIndex >= 0)
                        {
                            _audioUrlsChooser.SetVal(_audioUrlsChooser.choices[audioIndex]);
                        }
                        else
                        {
                            _audioUrlsChooser.SetVal("");
                        }
                    }
                });

                _uiContainerSimple.Add(_parentScript.CreateScrollablePopup(_audioUrlsChooser, false));

                UIDynamicButton testAudioFileButton = _uiContainerSimple.Add(_parentScript.CreateButton("Test Audio File", false));
                testAudioFileButton.button.onClick.AddListener(() =>
                {
                    if (!string.IsNullOrEmpty(_audioUrlsChooser.val))
                    {
                        int audioIndex = _audioUrlsChooser.choices.IndexOf(_audioUrlsChooser.val);
                        //PlayAudioClip(audioIndex);
                        NamedAudioClip audioClip = _audioClips?[audioIndex];
                        if (audioClip != null)
                        {
                            URLAudioClipManager.singleton.TestClip(audioClip);
                        }
                    }
                });

                _uiContainerSimple.Add(_parentScript.CreateSpacer(false)).height = 490f;
            }

            private void initAudioAdvanced()
            {
                // placeholder
            }

            private void initHelpHeader(UIDynamicContainer uiContainer, string header, string helpText)
            {
                UIDynamicButton left = uiContainer.Add(_parentScript.CreateButton($"<b>{header}</b>", false));
                UIDynamicButton right = uiContainer.Add(_parentScript.CreateButton("<b>?</b>", true));
                left.buttonColor = _parentScript.Cyan;
                right.buttonColor = _parentScript.Cyan;
                left.button.onClick.AddListener(() => SlapStuff.ShowHelpDialog(header, helpText));
                right.button.onClick.AddListener(() => SlapStuff.ShowHelpDialog(header, helpText));
                //uiContainer.Add(_parentScript.CreateSpacer(true)).height = 50f;

                uiContainer.Add(_parentScript.CreateSpacer(false)).height = 10f;
                uiContainer.Add(_parentScript.CreateSpacer(true)).height = 10f;
            }

            private void initBasicSettingsHelp()
            {
                initHelpHeader(_uiContainerSimple, "Basic Settings",
                    "- [Velocity] settings control when slaps are triggered and also at which intensity, because:\n" +
                    "\n- The [Volume] and [Pitch] Range will be interpolated between the [Velocity] range.\n" +
                    "\n\n** Advanced Options Only **\n" +
                    "\n- The [Interpolation Exponent] controls how pitch and volume change with respect to the velocity.\n" +
                    "  * A value of 1 means a simple linear increase with velocity.\n" +
                    "  * A value > 1 means that volume/pitch will increase slower at the start and faster towards the end.\n" +
                    "  * A value < 1 means that volume/pitch will increase faster at the start and slower towards the end.\n" +
                    "\n- [Use Frontal Velocity Fraction] to avoid triggering slaps when your are just grazing your target.\n" +
                    "\n- A [Cutoff at End Velocity] will cause that no slaps are triggered when the measured velocity is higher" +
                    " than the [(End) Velocity]. This can be useful, if you want handle light and hard slaps separately in two controllers.\n"
                );
            }

            private void initCollidersHelp()
            {
                initHelpHeader(_uiContainerSimple, "Colliders (This Atom)",
                    "Here you can choose the colliders from the containing atom of this plugin that shall trigger slaps.\n" +
                    "\nJust select them one by one from the drop down menu and click [Add Collider].\n" +
                    "\nBy default you can only see a smaller choice of potentially interesting colliders to add.\n" +
                    "\nNote that not all additional colliders that will show up when you enable [Show All Colliders]," +
                    " will actually be able to trigger slaps.\n"
                );
            }

            private void initConstraintsHelp()
            {
                initHelpHeader(_uiContainerAdvanced, "Collision Constraints",
                    "Here you can further constrain the conditions under which slaps shall be triggered.\n" +
                    "\n- The [Target Collider Filter]s are a list of words from which at least one needs to be contained" +
                    " in the name of the other collider you've hit. The comparison is case invariant." +
                    " The option [Enable Debug Log] in the main menu can help you find out which colliders you are actually hitting.\n" +
                    "To add a new filter, type in something like \"glute\" or \"thigh\" in the \"Filter...\" part of the drop down menu" +
                    " and click [Add Filter].\n" +
                    "\n- A minimum [Timeout] between triggered slaps above 0 (seconds) is highly recommended, because actually a lot of collisions" +
                    " are happening in a short amount time once you hit or touch something.\n" +
                    "\n- The [Target Atom Type] is a weaker constraint than the [Target Collider Filter]." +
                    " It let's choose the type of the Atoms as whole that you can hit (e.g. Persons).\n" +
                    "\n- [Prevent Self Collision] means that no slaps will be triggered if you hit any collider of the containing atom of this plugin.\n"
                );
            }

            private void initAudioHelp()
            {
                initHelpHeader(_uiContainerSimple, "Audio Files",
                    "The audio files added will be played at random for each triggered slap.\n" +
                    "\nJust click [Add Audio File] and browse in either local files or AddonPackages.\n" +
                    "\nTo obtain some audio files to get started, try to locate the Addon Package ToumeiHitsuji.SlapStuffAudioPack.\n" +
                    "\nNote that you can enable \"Keep Open\" in the file browser to add multiple files more easily.\n"
                );
            }

            public void ShowUi(bool advancedOptions = false)
            {
                _uiContainerSimple.Show();
                if (advancedOptions)
                {
                    _uiContainerAdvanced.Show();
                }
                else
                {
                    _uiContainerAdvanced.Hide();
                }
            }
            public void HideUi()
            {
                _uiContainerSimple.Hide();
                _uiContainerAdvanced.Hide();
            }

            private void syncColliderChoices()
            {
                List<string> choices = _colliderListJSON.valNoCallback.ToList(); // keep a copy of the added items on top
                if (choices.Count > 0)
                {
                    choices.Add("");                                             // followed by an empty string
                }
                if (_showAllCollidersJSON.valNoCallback)
                {
                    choices.AddRange(from rg in _parentScript.containingAtom.rigidbodies where rg.detectCollisions select rg.name);
                }
                else
                {
                    List<string> rgNames = SuperController.singleton.GetRigidbodyNamesInAtom(_parentScript.containingAtom.uid);
                    choices.AddRange(from name in rgNames where !name.ToLower().Contains("control") select name);
                }
                _colliderChooser.choices = choices;
            }

            private void clearColliders()
            {
                foreach (SlapCollisionHandler collisionHandler in _collisionHandlers)
                {
                    UnityEngine.Object.Destroy(collisionHandler);
                }
                _collisionHandlers.Clear();
                _colliderListTextJSON.valNoCallback = "";
            }

            protected void syncColliderListText()
            {
                _colliderListTextJSON.valNoCallback = "<b>Added Colliders (This Atom):</b>\n\n" +
                    string.Join("\n", (from name in _colliderListJSON.valNoCallback select $"- {name}").ToArray());
                syncColliderChoices();
            }

            protected void syncTargetColliderFilterChoices()
            {
                _targetColliderFilterChooser.choices = _targetColliderFilterJSON.valNoCallback.ToList();
            }

            private void parseColliders()
            {
                clearColliders();
                foreach (string name in _colliderListJSON.valNoCallback)
                {
                    addCollider(name, false);
                }
                syncColliderListText();
            }

            private void addCollider(string name, bool syncText=true)
            {
                Rigidbody collider = _parentScript.containingAtom.rigidbodies.First(rg => rg.name == name);
                collider.gameObject.AddComponent<SlapCollisionHandler>();
                SlapCollisionHandler collisionHandler = collider.GetComponent<SlapCollisionHandler>();
                collisionHandler.SlapController = this;
                collisionHandler.Collider = collider;
                _collisionHandlers.Add(collisionHandler);
                if (syncText)
                {
                    syncColliderListText();
                }
            }

            private void removeCollider(int index)
            {
                UnityEngine.Object.Destroy(_collisionHandlers[index]);
                _collisionHandlers.RemoveAt(index);
                syncColliderListText();
            }

            protected void onAtomRename(string oldid, string newid)
            {
                if (oldid == _audioSourceUidJSON.valNoCallback)
                {
                    _audioSourceUidJSON.valNoCallback = newid;
                }
            }

            protected void ensureAudioSource()
            {
                // might also have been renamed or deleted... best to create another one
                if (_audioSource == null || SuperController.singleton.GetAtomByUid(_audioSourceUidJSON.val) != _audioSource)
                {
                    _audioSource = null;
                    if (!_creatingAudioSource)
                    {
                        _parentScript.StartCoroutine(createAudioSourceCoroutine());
                    }
                }
            }

            private IEnumerator createAudioSourceCoroutine()
            {
                if (!_creatingAudioSource)
                {
                    _creatingAudioSource = true;
                    yield return null;
                    string atomUID = _audioSourceUidJSON.val;
                    _audioSource = SuperController.singleton.GetAtomByUid(atomUID);
                    if (_audioSource == null)
                    {
                        // if it didn't exist, make sure to use another (non default) name to ensure storing to json
                        atomUID = RandomAudioSourceName;
                        Logger.LogDebugMessage($"Creating AudioSource: {atomUID}.");
                        yield return SuperController.singleton.AddAtomByType("AudioSource", atomUID, false, true);
                        _audioSource = SuperController.singleton.GetAtomByUid(atomUID);
                        if (_audioSource == null)
                        {
                            //yield return null;
                            _audioSource = SuperController.singleton.GetSelectedAtom();
                        }
                        SuperController.singleton.SelectController(_parentScript.containingAtom.freeControllers[0]);
                    }
                    if (_audioSource != null)
                    {
                        _audioSource.hidden = true;
                        // having no AudioSourceControl can not only happen when it's the wrong object,
                        // but also when the AudioSource is turned off. So this might actually be intentional.
                        // Let's keep this sanity check commented out (when playing audio it is checked again... nothing bad should happen)
                        /*AudioSourceControl audioSourceControl = _audioSource.GetComponentInChildren<AudioSourceControl>();
                        if (audioSourceControl == null) // object is no audio source? -> create new one with different name next time
                        {
                            Logger.LogError($"{_audioSource.uid} has no AudioSourceControl");
                            _audioSourceUidJSON.val = RandomAudioSourceName;
                            _audioSource = null;
                        }
                        else*/
                        {
                            _audioSourceUidJSON.val = _audioSource.uid;
                            Logger.LogMessage($"Created AudioSource: {_audioSource.uid}. Do not delete manually!");
                            try
                            {
                                SuperController.singleton.onAtomUIDRenameHandlers -= onAtomRename;
                            }
                            catch (Exception)
                            {
                                // NOP - just assure not adding things twice
                            }
                            SuperController.singleton.onAtomUIDRenameHandlers += onAtomRename;
                        }
                    }
                    _creatingAudioSource = false;
                }
            }

            protected NamedAudioClip audioClipFromUrl(string url)
            {
                url = SuperController.singleton.NormalizeLoadPath(url);
                NamedAudioClip audioClip = URLAudioClipManager.singleton.GetClip(url);
                if (audioClip == null)
                {
                    URLAudioClip urlClip = URLAudioClipManager.singleton.QueueClip(SuperController.singleton.NormalizeMediaPath(url));
                    if (urlClip != null)
                    {
                         audioClip = URLAudioClipManager.singleton.GetClip(urlClip.uid);
                    }
                }
                return audioClip;
            }

            public static string Basename(string path)
            {
                return path.Substring(path.LastIndexOfAny(new char[] { '/', '\\' }) + 1);
            }

            protected void syncAudioUrlChoices()
            {
                List<string> choices = new List<string>();
                int i = 0;
                foreach (NamedAudioClip audioClip in _audioClips)
                {
                    choices.Add($"{i}: {audioClip.displayName}");
                    i++;
                }
                _audioUrlsChooser.choices = choices;
                if (string.IsNullOrEmpty(_audioUrlsChooser.val) && _audioUrlsChooser.choices.Count > 0)
                {
                    _audioUrlsChooser.valNoCallback = _audioUrlsChooser.choices[0];
                }
            }

            protected void syncAudioUrlValidity()
            {
                // remove null
                for (int i = _audioClips.Count - 1; i >= 0; --i)
                {
                    if (_audioClips[i] == null)
                    {
                        Logger.LogDebugError($"Invalid audio url: {_audioUrlListJSON.valNoCallback[i]}");
                        _audioUrlListJSON.RemoveAtNoCallback(i);
                        _audioClips.RemoveAt(i);
                    }
                }
            }

            protected void syncAudioUrlsText()
            {
                List<string> urls = _audioUrlListJSON.valNoCallback;
                if (urls.Count != _audioClips.Count)
                {
                    Logger.LogError("Internal error when syncing audio urls.");
                }
                _audioUrlListTextJSON.valNoCallback = "<b>Added Audio Files:</b>\n\n";
                if (urls.Count > 0)
                {
                    string[] audioUrlsWithIndexes =
                        (from urlAndIndex in urls.Select((u, i) => new {Url = u, Index = i}) select $"{urlAndIndex.Index}: {urlAndIndex.Url}").ToArray();
                    _audioUrlListTextJSON.valNoCallback += string.Join("\n\n", audioUrlsWithIndexes);
                }
            }
            private void clearAudioClips()
            {
                if (!_parentScript.KeepAudioInScene)
                {
                    foreach (NamedAudioClip audioClip in _audioClips)
                    {
                        URLAudioClipManager.singleton.RemoveClip(audioClip);
                    }
                }
                _audioClips.Clear();
            }

            private void parseAudioUrls()
            {
                clearAudioClips();
                foreach (string url in _audioUrlListJSON.valNoCallback)
                {
                    addAudioFile(url, false);
                }
                syncAudioUrlValidity();
                syncAudioUrlsText();
                syncAudioUrlChoices();
            }

            private void addAudioFile(string url, bool sync=true)
            {
                // create audio source only when the first file is added
                ensureAudioSource();

                NamedAudioClip audioClip = audioClipFromUrl(url);
                _audioClips.Add(audioClip);

                if (sync)
                {
                    syncAudioUrlValidity();
                    syncAudioUrlsText();
                    syncAudioUrlChoices();
                }
            }

            private void removeAudioFile(int index)
            {
                if (!_parentScript.KeepAudioInScene)
                {
                    URLAudioClipManager.singleton.RemoveClip(_audioClips[index]);
                }
                _audioClips.RemoveAt(index);
                syncAudioUrlsText();
                syncAudioUrlChoices();
            }

            private void removeAudioSource()
            {
                if (_audioSource != null && SuperController.singleton.GetAtomByUid(_audioSource.uid) != null)
                {
                    SuperController.singleton.onAtomUIDRenameHandlers -= onAtomRename;
                    SuperController.singleton.RemoveAtom(_audioSource, true);
                }
                _audioSource = null;
                //Destroy(_audioSource);
            }

            public void PlayAudioClip(int index, float volume, float pitch, Vector3 position)
            {
                Atom audioSource = SlapAudioSource;
                if (audioSource == null)
                {
                    Logger.LogError("There is no audio source.");
                    return;
                }
                AudioSourceControl audioSourceControl = audioSource.GetComponentInChildren<AudioSourceControl>();
                if (audioSourceControl == null)
                {
                    // This can actually happen if the audio source is turned off, which might be intentional?
                    Logger.LogDebugError($"{audioSource.uid} has no AudioSourceControl.");
                    return;
                }
                audioSource.mainController.control.transform.position = position;
                NamedAudioClip audioClip = _audioClips?[index];
                if (audioClip != null)
                {
                    audioSourceControl.volume = volume;
                    audioSourceControl.pitch = pitch;
                    audioSourceControl.PlayNowClearQueue(audioClip);
                }
                else
                {
                    Logger.LogError("Invalid audio clip.");
                }
            }

            public void PlayAudioClip(int index, float volume=1f, float pitch=1f)
            {
                // 50cm in front of camera
                Vector3 position = SuperController.singleton.lookCamera.transform.position +
                                   SuperController.singleton.lookCamera.transform.forward * 0.5f;
                PlayAudioClip(index, volume, pitch, position);
            }

            public void Destroy()
            {
                try
                {
                    clearColliders();
                    removeAudioSource();
                    clearAudioClips();
                    _uiContainerSimple.Clear();
                    _uiContainerAdvanced.Clear();
                    Logger.LogDebugMessage($"{nameof(SlapCollisionController)} destroyed");
                }
                catch (Exception e)
                {
                    Logger.LogError($"{nameof(SlapCollisionController)}.{nameof(Destroy)}: {e}");
                }
            }

            public void Disable()
            {
                try
                {
                    foreach (SlapCollisionHandler collisionHandler in _collisionHandlers)
                    {
                        collisionHandler.enabled = false;
                    }
                    Logger.LogDebugMessage($"{nameof(SlapCollisionController)} disabled");
                }
                catch (Exception e)
                {
                    Logger.LogError($"{nameof(SlapCollisionController)}.{nameof(Destroy)}: {e}");
                }
            }

            public void Enable()
            {
                try
                {
                    foreach (SlapCollisionHandler collisionHandler in _collisionHandlers)
                    {
                        collisionHandler.enabled = true;
                    }
                    Logger.LogDebugMessage($"{nameof(SlapCollisionController)} disabled");
                }
                catch (Exception e)
                {
                    Logger.LogError($"{nameof(SlapCollisionController)}.{nameof(Destroy)}: {e}");
                }
            }

        }

    }

}
