using System;
using System.Collections.Generic;
using UnityEngine;
using System.Linq;
using System.Collections;

namespace ToumeiHitsuji
{

    public partial class SlapStuff
    {
        public class SlapCollisionHandler : MonoBehaviour
        {
            public SlapCollisionController SlapController = null;
            public Rigidbody Collider = null;

            void OnCollisionEnter(Collision collision)
            {
                try
                {
                    if (SlapController != null && enabled)
                    {
                        SlapController.ReactOnCollision(collision, this);
                    }
                }
                catch (Exception e)
                {
                    Logger.LogError($"{nameof(SlapCollisionHandler)}.{nameof(OnCollisionEnter)}: {e}");
                }
            }
            public void OnDestroy()
            {
                try
                {
                    Logger.LogDebugMessage($"{nameof(SlapCollisionHandler)} destroyed");
                }
                catch (Exception e)
                {
                    Logger.LogError($"{nameof(SlapCollisionHandler)}.{nameof(OnDestroy)}: {e}");
                }
            }

            public void OnEnable()
            {
                try
                {
                    Logger.LogDebugMessage($"{nameof(SlapCollisionHandler)} enabled");
                }
                catch (Exception e)
                {
                    Logger.LogError($"{nameof(SlapCollisionHandler)}.{nameof(OnEnable)}: {e}");
                }
            }

            public void OnDisable()
            {
                try
                {
                    Logger.LogDebugMessage($"{nameof(SlapCollisionHandler)} disabled");
                }
                catch (Exception e)
                {
                    Logger.LogError($"{nameof(SlapCollisionHandler)}.{nameof(OnDisable)}: {e}");
                }
            }
        }

        public partial class SlapCollisionController
        {
            private bool _isInTimeOut = false;

            private IEnumerator timeOutCoroutine()
            {
                _isInTimeOut = true;
                yield return new WaitForSeconds(_timeOutJSON.val);
                _isInTimeOut = false;
            }

            private bool isValidCollisionTarget(Collision collision)
            {
                ForceReceiver forceReceiver = collision.gameObject.GetComponentInParent<ForceReceiver>();
                Atom collidingAtom = forceReceiver?.containingAtom;
                if (collidingAtom == null)
                {
                    // this does not always return the atom the game object belongs to,
                    // if the atom is parented to another atom, it will return this... (still better than nothing)
                    collidingAtom = collision.gameObject.GetComponentInParent<Atom>();
                }
                string atomType = "";
                if (collidingAtom != null)
                {
                    if (_preventSelfCollisionJSON.val & collidingAtom == _parentScript.containingAtom)
                    {
                        return false;
                    }
                    atomType = collidingAtom.type;
                }
                if (_atomTypeChooser.val != "All" && atomType != _atomTypeChooser.val)
                {
                    return false;
                }
                if (_targetColliderFilterJSON.valNoCallback.Count > 0)
                {
                    foreach(string filter in _targetColliderFilterJSON.valNoCallback)
                    {
                        // case invariant contains
                        if (collision.collider.name.IndexOf(filter, StringComparison.OrdinalIgnoreCase) >= 0)
                        {
                            return true;
                        }
                    }
                    return false;
                }
                return true;
            }

            public void ReactOnCollision(Collision collision, SlapCollisionHandler collisionHandler)
            {
                if (_isInTimeOut)
                {
                    //Logger.LogDebugMessage("Timeout");
                    return;
                }
                Logger.LogDebugMessage($"{collisionHandler.Collider.name} --> {collision.collider.name}");
                float velocity = collision.relativeVelocity.magnitude;
                Logger.LogDebugMessage($"Velocity: {velocity}");
                if (velocity < _velocityStartJSON.val || collision.rigidbody == null)
                {
                    return;
                }
                if (_audioClips == null || _audioClips.Count <= 0)
                {
                    Logger.LogDebugMessage("There are no audio clips.");
                    return;
                }

                Atom parentAtom = collision.gameObject.GetComponentInParent<Atom>();
                if (!isValidCollisionTarget(collision))
                {
                    return;
                }

                float frontalVelocity = 0f;
                Vector3 audioOrigin = new Vector3(0f, 0f, 0f);
                foreach (ContactPoint contact in collision.contacts)
                {
                    audioOrigin += contact.point;
                    frontalVelocity += Mathf.Abs(Vector3.Dot(contact.normal, collision.relativeVelocity));
                }
                audioOrigin /= (float)collision.contacts.Count();
                frontalVelocity /= (float)collision.contacts.Count();
                Logger.LogDebugMessage($"Frontal Velocity: {frontalVelocity}");
                if (_useFrontalVelocityJSON.val)
                {
                    velocity = frontalVelocity;
                }
                if (velocity < _velocityStartJSON.val || (_cutoffAtEndVelocityJSON.val && velocity > _velocityEndJSON.val))
                {
                    return;
                }
                _parentScript.StartCoroutine(timeOutCoroutine());

                float normalizedVelocity = Mathf.Clamp((velocity - _velocityStartJSON.val) / (_velocityEndJSON.val - _velocityStartJSON.val), 0f, 1f);
                normalizedVelocity = Mathf.Pow(normalizedVelocity, _interpolationExponentJSON.val);
                PlayAudioClip(
                    UnityEngine.Random.Range(0, _audioClips.Count),
                    Mathf.Lerp(_volumeStartJSON.val, _volumeEndJSON.val, normalizedVelocity),
                    Mathf.Lerp(_pitchStartJSON.val, _pitchEndJSON.val, normalizedVelocity),
                    audioOrigin);
            }
        }
    }

}
