using System;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using UnityEngine;
using System.Linq;
using System.Collections;
using SimpleJSON;
using uFileBrowser;
using MVR.FileManagementSecure;

namespace ToumeiHitsuji
{

    public partial class SlapStuff : MVRScript
    {
        public string VersionInfo { get { return "1.1.0"; } }
        public string CreatorName { get { return "ToumeiHitsuji"; } }
        private UIDynamicContainer _mainMenu;
        private UIDynamicButton _addControllerButton;
        private UIDynamicButton _removeControllerButton;
        private UIDynamicButton _addPresetButton;
        private UIDynamicButton _saveControllerPresetButton;
        private JSONStorableStringChooser _menuChooser = new JSONStorableStringChooser("Menu Choice", new List<string>{"<b>Main Menu</b>"}, "<b>Main Menu</b>", "Menu");
        private string _addControllerChoice = "<i>+ New Controller</i>";
        private string _addPresetChoice = "<i>+ From Preset</i>";
        private JSONStorableBool _advancedOptionsJSON = new JSONStorableBool("<b>Show Advanced Options</b>", false);
        private JSONStorableBool _keepAudioInSceneJSON = new JSONStorableBool("<b>Always Keep Audio in Scene</b>", true);
        private JSONStorableBool _enableDebugLogJSON = new JSONStorableBool("Enable Debug Log", false);
        private List<SlapCollisionController> _collisionControllers = new List<SlapCollisionController>();
        private int _currentMenuIndex = -1;
        private int _selectedMenuIndex = 0;
        public bool KeepAudioInScene { get { return _keepAudioInSceneJSON.val; } }
        public Color Green { get { return new Color(.5f, .96f, .5f); } }
        public Color Red { get { return new Color(.96f, .45f, .45f); } }
        public Color Blue { get { return new Color(.05f, .37f, 0.62f); } }
        public Color LightBlue { get { return new Color(.55f, .60f, 0.68f); } }
        public Color Yellow { get { return Color.yellow; } }
        public Color Cyan { get { return Color.cyan; } }

        public bool NeedsUiUpdate
        {
            get { return _currentMenuIndex != _selectedMenuIndex; }
            set { _currentMenuIndex = value ? -1 : _currentMenuIndex; }
        }

        public int CurrentControllerIndex
        {
            get { return _selectedMenuIndex - 1; }
        }

        private void hideAllMenus()
        {
            _mainMenu.Hide();
            foreach (SlapCollisionController cc in _collisionControllers)
            {
                cc.HideUi();
            }
        }

        public void Update()
        {
            if (NeedsUiUpdate)
            {
                // somehow this can't be done directly during init
                showMenu(_selectedMenuIndex);
            }
        }

        private void showMenu(int index)
        {
            hideAllMenus();
            _selectedMenuIndex = Mathf.Clamp(index, 0, _collisionControllers.Count());
            _currentMenuIndex = _selectedMenuIndex;
            if (_selectedMenuIndex == 0)
            {
                _mainMenu.Show();
                _removeControllerButton.gameObject.SetActive(false);
                _saveControllerPresetButton.gameObject.SetActive(false);
                _addControllerButton.gameObject.SetActive(true);
                _addPresetButton.gameObject.SetActive(true);
            }
            else
            {
                _addControllerButton.gameObject.SetActive(false);
                _addPresetButton.gameObject.SetActive(false);
                _removeControllerButton.gameObject.SetActive(true);
                _saveControllerPresetButton.gameObject.SetActive(true);
                _collisionControllers[CurrentControllerIndex].ShowUi(_advancedOptionsJSON.val);
            }
            _menuChooser.valNoCallback = _menuChooser.choices[_selectedMenuIndex];
        }

        private void syncMenuChoices()
        {
            List<string> choices = new List<string>{"<b>Main Menu</b>"};
            for (int i = 0; i < _collisionControllers.Count(); ++i)
            {
                choices.Add($"<b>Controller {i+1}</b>");
            }
            choices.Add(_addControllerChoice);
            choices.Add(_addPresetChoice);
            _menuChooser.choices = choices;
        }

        private void addCollisionController()
        {
            _collisionControllers.Add(new SlapCollisionController(this));
            syncMenuChoices();
            _selectedMenuIndex = _collisionControllers.Count();
        }

        private void addCollisionControllerPreset()
        {
            PluginFileDialog("Load Preset", "json", "Presets", false, (string path) => {
                JSONClass jc = SuperController.singleton.LoadJSON(path).AsObject;
                loadCollisionControllers(jc, true, true, true);
            });
        }

        private void removeCollisionController(int index)
        {
            if (index < 0 || index >= _collisionControllers.Count())
            {
                return;
            }
            _collisionControllers[index].Destroy();
            _collisionControllers.RemoveAt(index);
            syncMenuChoices();
            NeedsUiUpdate = true;
        }

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            return GetJSON(-1, false, includePhysical, includeAppearance, forceStore);
        }

        public JSONClass GetJSON(int controllerIndex, bool asPreset = false, bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);

            // Always update version info! (it might have been overwritten by a loaded preset)
            jc["Version"] = VersionInfo;

            JSONArray ccJsonArray = new JSONArray();

            if (controllerIndex < 0 | controllerIndex >= _collisionControllers.Count)
            {
                foreach(SlapCollisionController cc in _collisionControllers)
                {
                    ccJsonArray.Add(cc.GetJSON(includePhysical, includeAppearance, forceStore));
                }
            }
            else
            {
                ccJsonArray.Add(_collisionControllers[controllerIndex].GetJSON(includePhysical, includeAppearance, forceStore));
            }

            if (asPreset)
            {
                foreach (JSONNode jn in ccJsonArray)
                {
                    // these names should never be saved in presets, but they are very important in scene saves
                    jn.Remove("Collision AudioSource");
                }
            }

            jc["Collision Controllers"] = ccJsonArray;
            return jc;
        }

        public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
        {
            // jc.Count > 0 is to avoid an "empty restore" when someone just loads a look or resets the pose.
            // Apart from why the hell this triggers a restore in the first place, it would usually not be a problem,
            // because it doesn't set any values. But my custom restore routine involves a prior clean up,
            // and when there is nothing to restore, it just stays cleaned up...
            if (jc.Count > 0)
            {
                base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
                StartCoroutine(restoreCoroutine(jc, restorePhysical, restoreAppearance, setMissingToDefault));
            }
        }

        private IEnumerator restoreCoroutine(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
        {
            clearCollisionControllers();
            // seems like it is better to have a frame delay after cleanup, otherwise loading a preset sometimes doesn't work until reloading the scene
            yield return null;
            loadCollisionControllers(jc, restorePhysical, restoreAppearance, setMissingToDefault);
        }

        private void loadCollisionControllers(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
        {
            foreach (JSONNode jn in jc["Collision Controllers"].AsArray)
            {
                addCollisionController();
                SlapCollisionController cc = _collisionControllers.Last();
                // guess I have to call both restores??
                cc.RestoreFromJSON(jn.AsObject, restorePhysical, restoreAppearance, setMissingToDefault);
                cc.LateRestoreFromJSON(jn.AsObject, restorePhysical, restoreAppearance, setMissingToDefault);
            }
        }

        public override void Init()
        {
            try
            {
                // Version info for potential future compatibility checks of presets
                JSONStorableString versionInfoJSON = new JSONStorableString("Version", "");
                versionInfoJSON.storeType = JSONStorableParam.StoreType.Full;
                RegisterString(versionInfoJSON);
                // Set as a different value than starting value to make sure it is stored.
                // This is also important to make sure that at least one thing from the main class is there to store,
                // otherwise restore might not be called and collision controllers won't be restored on load.
                // Don't know if there is a less roundabout way to ensure correct load...
                versionInfoJSON.val = VersionInfo;

                _mainMenu = new UIDynamicContainer(this);

                _menuChooser.storeType = JSONStorableParam.StoreType.Full;
                UIDynamicPopup menuPopup = CreateScrollablePopup(_menuChooser, false);
                menuPopup.popup.showSlider = false;
                menuPopup.height = 115f;
                menuPopup.popup.onOpenPopupHandlers += syncMenuChoices;
                _menuChooser.setCallbackFunction = (string choice) =>
                {
                    if (choice == _addControllerChoice)
                    {
                        addCollisionController();
                    }
                    else if (choice == _addPresetChoice)
                    {
                        addCollisionControllerPreset();
                    }
                    else
                    {
                        _selectedMenuIndex = _menuChooser.choices.IndexOf(choice);
                    }
                };

                _addControllerButton = CreateButton("<b>Add New Controller</b>", true);
                _addControllerButton.buttonColor = Green;
                _addControllerButton.button.onClick.AddListener(addCollisionController);

                _addPresetButton = CreateButton("<b>Add Controller From Preset</b>", true);
                _addPresetButton.buttonColor = Green;
                _addPresetButton.button.onClick.AddListener(addCollisionControllerPreset);

                _removeControllerButton = CreateButton("<b>Remove Controller</b>", true);
                _removeControllerButton.buttonColor = Red;
                _removeControllerButton.button.onClick.AddListener(()=>
                {
                    removeCollisionController(_selectedMenuIndex - 1);
                });

                _saveControllerPresetButton = CreateButton("<b>Save Controller As Preset</b>", true);
                _saveControllerPresetButton.buttonColor = Green;
                _saveControllerPresetButton.button.onClick.AddListener(()=>
                {
                    PluginFileDialog("Save Preset", "json", "Presets", true, (string path) => {
                        JSONClass jc = GetJSON(CurrentControllerIndex, true, true, true, true);
                        SuperController.singleton.SaveJSON(jc, path);
                        Logger.LogDebugMessage($"Saved preset to: {path}");
                    });
                });

                CreateSpacer(false).height = 80f;
                CreateSpacer(true).height = 80f;

                UIDynamicButton loadPresetButton = _mainMenu.Add(CreateButton("Load Preset", false));
                loadPresetButton.button.onClick.AddListener(()=>
                {
                    PluginFileDialog("Load Preset", "json", "Presets", false, (string path) => {
                        JSONClass jc = SuperController.singleton.LoadJSON(path).AsObject;
                        RestoreFromJSON(jc, true, true, null, true);
                    });
                });

                UIDynamicButton savePresetButton = _mainMenu.Add(CreateButton("Save Preset", false));
                savePresetButton.button.onClick.AddListener(()=>
                {
                    if (_collisionControllers.Count == 0)
                    {
                        Logger.LogMessage("There is nothing to save, yet.");
                        return;
                    }
                    PluginFileDialog("Save Preset", "json", "Presets", true, (string path) => {
                        JSONClass jc = GetJSON(-1, true, true, true, true);
                        SuperController.singleton.SaveJSON(jc, path);
                        Logger.LogDebugMessage($"Saved preset to: {path}");
                    });
                });

                _mainMenu.Add(CreateSpacer(false)).height = 145f;


                JSONStorableString presetInfoJSON = new JSONStorableString("Info",
                    "<size=30><b>Quick Start Tip:</b></size>\n" +
                    "Load a preset from this plugin's var package. As for the manual setup...");
                _mainMenu.Add(CreateTextField(presetInfoJSON, true)).height = 115f;

                UIDynamicButton openHowToButton = _mainMenu.Add(CreateButton("<b>...Blue Buttons Show Help!</b>", true));
                openHowToButton.buttonColor = Cyan;
                openHowToButton.button.onClick.AddListener(HowToDialog);

                _mainMenu.Add(CreateSpacer(true)).height = 80f;

                _keepAudioInSceneJSON.storeType = JSONStorableParam.StoreType.Full;
                RegisterBool(_keepAudioInSceneJSON);
                UIDynamicToggle keepAudioToggle = _mainMenu.Add(CreateToggle(_keepAudioInSceneJSON, false));
                keepAudioToggle.backgroundColor = Green;
                JSONStorableString keepAudioInfoJSON = new JSONStorableString("Info",
                    "This option will keep audio clips in scene audio even if you remove " +
                    "them from within this plugin or if the plugin itself is removed. " +
                    "This is important if any other source still wants to use the clips " +
                    "(it may even be just another collision controller inside the same plugin instance). " +
                    "<b>Only uncheck if you know what you are doing.</b>");
                UIDynamicTextField keepAudioInfoTextfield =  _mainMenu.Add(CreateTextField(keepAudioInfoJSON, false));
                keepAudioInfoTextfield.height = 300f;
                _keepAudioInSceneJSON.setCallbackFunction = (bool isChecked) =>
                {
                    keepAudioToggle.backgroundColor = isChecked ? Green : Red;
                };

                UIDynamicToggle enableDebugLogToggle = _mainMenu.Add(CreateToggle(_enableDebugLogJSON, true));
                enableDebugLogToggle.backgroundColor = Cyan;
                _enableDebugLogJSON.setCallbackFunction = (bool isChecked) =>
                {
                    Logger.DebugEnabled = isChecked;
                };
                JSONStorableString enableDebugLogInfoJSON = new JSONStorableString("Info",
                    "The debug log might help you find the right velocity thresholds. " +
                    "It also tells you what colliders you hit.");
                _mainMenu.Add(CreateTextField(enableDebugLogInfoJSON, true)).height = 300f;

                _mainMenu.Add(CreateSpacer(true)).height = 80f;

                _advancedOptionsJSON.storeType = JSONStorableParam.StoreType.Full;
                //RegisterBool(_advancedOptionsJSON);
                UIDynamicToggle advancedOptionsToggle = CreateToggle(_advancedOptionsJSON, true);
                advancedOptionsToggle.backgroundColor = Yellow;
                _mainMenu.Add(advancedOptionsToggle);

                Logger.LogDebugMessage($"{nameof(SlapStuff)} initialized");
            }
            catch (Exception e)
            {
                Logger.LogError($"{nameof(SlapStuff)}.{nameof(Init)}: {e}");
            }
        }

        public void PluginFileDialog(string title, string fileFormat, string subDir, bool saveDialog, FileBrowserCallback callback)
        {
            FileBrowser fb = SuperController.singleton.fileBrowserUI;
            string localPath = $"Saves\\PluginData\\{CreatorName}\\{nameof(SlapStuff)}";
            if (!string.IsNullOrEmpty(subDir))
            {
                localPath = $"{localPath}\\{subDir}";
            }
            if (!FileManagerSecure.DirectoryExists(localPath))
            {
                Logger.LogDebugMessage($"Creating directory: {localPath}");
                FileManagerSecure.CreateDirectory(localPath);
            }

            fb.fileRemovePrefix = null;
            fb.hideExtension = false;
            fb.keepOpen = false;
            fb.showDirs = true;
            fb.shortCuts = new List<ShortCut>{
                new ShortCut{path = localPath, displayName="PluginData Local", isLatest = true},
                new ShortCut{path = "Saves", displayName="Saves", isLatest = true},
                new ShortCut{path = "Custom", displayName="Custom", isLatest = true}
                };
            fb.defaultPath = localPath;
            if (!saveDialog)
            {
                string pluginPath = "";
                try
                {
                    string pluginRoot = manager.GetJSON()["plugins"][storeId.Substring(0, storeId.IndexOf('_'))].Value;
                    Match pluginRootMatch = new Regex($"\\A{CreatorName}" + @"\.\w+\.\d+:[/\\]").Match(pluginRoot);
                    if (pluginRootMatch.Success)
                    {
                        pluginRoot = pluginRootMatch.Value.Replace('/', '\\');
                        pluginPath = pluginRoot + localPath;
                    }
                }
                catch (Exception e)
                {
                    Logger.LogError($"{nameof(SlapStuff)}.{nameof(PluginFileDialog)}: {e}");
                }
                if (!string.IsNullOrEmpty(pluginPath) && FileManagerSecure.DirectoryExists(pluginPath))
                {
                    fb.defaultPath = pluginPath;
                    fb.shortCuts.Insert(0, new ShortCut{path = pluginPath, displayName="PluginData Package", isLatest = true});
                }
                fb.shortCuts.Add(new ShortCut{path = "AddonPackages", displayName="AddonPackages", isLatest = true});
            }
            // make sure shortcuts are updated
            fb.onlyShowLatest = false;
            fb.onlyShowLatest = true;

            fb.browseVarFilesAsDirectories = !saveDialog;
            fb.fileFormat = fileFormat;
            fb.SetTitle(title);
            fb.SetTextEntry(saveDialog);
            fb.Show((string path, bool didClose) =>
            {
                if (!string.IsNullOrEmpty(path) && didClose)
                {
                    if (!path.EndsWith($".{fileFormat}", StringComparison.OrdinalIgnoreCase))
                    {
                        path = $"{path}.{fileFormat}";
                    }

                    callback(FileManagerSecure.NormalizePath(path));
                }
            });
            if (saveDialog)
            {
                fb.ActivateFileNameField();
            }
        }

        public static void ShowHelpDialog(string header, string helpText)
        {
            SuperController.singleton.ClearMessages();
			SuperController.LogMessage($"-------------------- {header} --------------------\n\n{helpText}");
			SuperController.singleton.OpenMessageLogPanel();
        }

        public static void HowToDialog()
        {
			ShowHelpDialog("How To Use SlapStuff",
                "With SlapStuff you can configure collision sounds (like slaps) for any collider.\n" +
                "\nIn a single plugin instance you can configure multiple collision controllers for different combinations of colliders and sounds.\n" +
                "\nStart by adding a new controller on the top right of the main menu (or anytime from the menu drop down).\n" +
                "\nDon't worry, there will be more blue buttons on the way...\n" +
                "\nOnce you feel comfortable with setting things up, come back to the main menu and also try out the advanced options.\n"
            );
		}

        private void clearCollisionControllers()
        {
            if (_collisionControllers != null)
            {
                foreach(SlapCollisionController cc in _collisionControllers)
                {
                    cc.Destroy();
                }
                _collisionControllers.Clear();
            }
        }

        public void OnEnable()
        {
            try
            {
                if (_collisionControllers != null)
                {
                    foreach(SlapCollisionController cc in _collisionControllers)
                    {
                        cc.Enable();
                    }
                }
                Logger.LogDebugMessage($"{nameof(SlapStuff)} enabled");
            }
            catch (Exception e)
            {
                Logger.LogError($"{nameof(SlapStuff)}.{nameof(OnEnable)}: {e}");
            }
        }

        public void OnDisable()
        {
            try
            {
                if (_collisionControllers != null)
                {
                    foreach(SlapCollisionController cc in _collisionControllers)
                    {
                        cc.Disable();
                    }
                }
                Logger.LogDebugMessage($"{nameof(SlapStuff)} disabled");
            }
            catch (Exception e)
            {
                Logger.LogError($"{nameof(SlapStuff)}.{nameof(OnDisable)}: {e}");
            }
        }

        public void OnDestroy()
        {
            try
            {
                clearCollisionControllers();
                Logger.LogDebugMessage($"{nameof(SlapStuff)} destroyed");
            }
            catch (Exception e)
            {
                Logger.LogError($"{nameof(SlapStuff)}.{nameof(OnDestroy)}: {e}");
            }
        }
    }

}
